import mne

from ...data_browser import DataBrowser
from ..base_step import *
from ..base_preprocessor import BasePreprocessor
from .mne_step import MneStep

# This is the same as MnePreprocessorBasic, but without the default preprocessor
class MnePreprocessorAsIs(BasePreprocessor[mne.io.Raw]):
    """
    The MnePreprocessorAsIs class uses the loaded data as-is, which means that
    one should provide haemoglobin concentration already prepared in a previous
    preprocessing pipeline.    

    Use this preprocessor if you already cleaned your data and have haemoglobin concentration.

    This is the default preprocessor.
    """
    def __init__(self):
        super().__init__()
    
    def read_file(self, path:str, verbose:bool=False):
        if DataBrowser.is_path_fif(path):
            return mne.io.read_raw_fif(path, preload=True, verbose=verbose)

        if DataBrowser.is_path_nirx(path):
            return mne.io.read_raw_nirx(fname=path, preload=True, verbose=verbose)

        if DataBrowser.is_path_snirf(path):
            return mne.io.read_raw_snirf(path, preload=True, verbose=verbose)

        raise ValueError(f"No reader for file {path}")
    
    def run(self, raw: mne.io.Raw, verbose: bool = False) -> list[MneStep]:
        if verbose:
            print('Using MnePreprocessorAsIs, using raw data as already preprocessed')

        step = MneStep(raw, PREPROCESS_STEP_BASE_KEY, PREPROCESS_STEP_BASE_DESC)
        steps = [step]
        return steps
        



# This Cedalion Preprocessor is a sandbox and is provided as inspiration only for further development
# See https://openfnirs.org/training/cedalion/ for more info about the Cedalion project
#
# Here is how to have cedalion available as a python package for this project
#   poetry shell
#   pip install pint-xarray trimesh vtk pyvista strenum nibabel click trame
#   mkdir /path/where/you/want/to/install
#   cd /path/where/you/want/install
#   git clone https://github.com/ibs-lab/cedalion
#   cd cedalion
#   pip install -e .
#

try:
    import xarray as xr
    import cedalion
    import cedalion.dataclasses as cdc
    from cedalion.typing import NDTimeSeries
    import matplotlib.pyplot as plt

    from ...data_browser import DataBrowser
    from ..base_step import *
    from ..base_preprocessor import BasePreprocessor

    class CedalionStep(BaseStep[cdc.Recording]):
        @property
        def n_times(self):
            return len(self.obj['time'])
            
        @property
        def sfreq(self):
            # TODO We probably have this information directly
            return 1 / float(self.obj['time'][1] - self.obj['time'][0])
            
        @property
        def ch_names(self):
            # TODO learn how to use xarray efficiently
            # TODO maybe we should send channel and  wavelength
            return list(self.obj.channel.to_series())
            
        def plot(self, **_kwargs):

            # select which time series we work with
            ts = self.obj

            # Thanks to the xarray DataArray structure, we can easily select the data we want to plot
            # plot four channels and their stim markers
            f, ax = plt.subplots(len(self.ch_names), 1, sharex=True)
            for i, ch in enumerate(self.ch_names):
                ax[i].plot(ts.time, ts.sel(channel=ch, chromo="HbO"), "r-", label="HbO")
                ax[i].plot(ts.time, ts.sel(channel=ch, chromo="HbR"), "b-", label="HbR")
                ax[i].set_title(f"Ch. {ch}")
                # add stim markers using Cedalion's plot_stim_markers function
                #cedalion.plots.plot_stim_markers(ax[i], rec.stim, y=1)
                ax[i].set_ylabel(r"$\Delta$ c / uM")

            ax[0].legend(ncol=6)
            ax[3].set_label("time / s")
            ax[3].set_xlim(0,100)
            #plt.tight_layout()
            return f

    class CedalionPreprocessor(BasePreprocessor[cdc.Recording]):
        def read_file(self, path, verbose:bool=False) -> cdc.Recording:
            # TODO honor verbose arg
            if not DataBrowser.is_path_snirf(path):
                raise RuntimeError('Not implemented: only snirf file is supported for now')

            recordings = cedalion.io.read_snirf(path)
            rec: cdc.Recording  = recordings[0]

            # TODO we should force units only if they are not set
            rec['amp'] = rec['amp'].pint.dequantify().pint.quantify("V")
            rec['amp']['time'] = rec['amp']['time'].pint.dequantify().pint.quantify("second")
            return rec

        def run(self, rec, verbose: bool = False) -> list[CedalionStep]:
            # TODO honor verbose
            steps = []
            amp = rec['amp']
            steps.append(CedalionStep(amp, PREPROCESS_STEP_BASE_KEY, PREPROCESS_STEP_BASE_DESC))

            od = cedalion.nirs.int2od(amp)
            steps.append(CedalionStep(od, PREPROCESS_STEP_OD_KEY, PREPROCESS_STEP_OD_DESC))


            dpf = xr.DataArray(
                [6., 6.],
                dims="wavelength",
                coords={"wavelength" : [760., 850.]}) # TODO unhardcode wavelengts

            haemo = cedalion.nirs.beer_lambert(rec['amp'], rec.geo3d, dpf)
            steps.append(CedalionStep(haemo, PREPROCESS_STEP_HAEMO_KEY, PREPROCESS_STEP_HAEMO_DESC))

            return steps
            

except: 
    CedalionPreprocesStep = None
    CedalionPreprocessor = None


from typing import List
import os
from pathlib import Path
from zipfile import ZipFile

import pooch
import numpy as np
import scipy.io

from ..wavelet.pair_signals import PairSignals

DOWNLOADS_RELATIVE_PATH = os.path.join('data', 'NIRS', 'downloads')

class DataBrowser:
    absolute_root_path: str
    paths: List[str]

    def __init__(self):
        """
        The DataBrowser class allows to recursively list NIRS files in folders and detect file type

        It can also download demo dataset for convenience.
        """
        self.absolute_root_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        self.paths = [
            self.absolute_path(os.path.join('data')),
            self.absolute_path(os.path.join('data', 'fNIRS')),
            DOWNLOADS_RELATIVE_PATH,
        ]
    
    def absolute_path(self, relative_path:str) -> str:
        """
        Get absolute path from relative path

        Args:
            relative_path (str): a path relative to the project

        Returns:
            str: the absolute path on disk
        """
        return os.path.join(self.absolute_root_path, relative_path)

    def add_source(self, path:str):
        """
        Add a folder to look for NIRS files

        Args:
            path (str): absolute path to add

        Returns:
            self: the DataBrowser object itself. Useful for chaining operations
        """
        self.paths.append(path)
        return self

    @staticmethod
    def is_path_nirx(path:str) -> bool:
        return os.path.isfile(path) and path.endswith('.hdr')

    @staticmethod
    def is_path_fif(path:str) -> bool:
        return os.path.isfile(path) and path.endswith('.fif')

    @staticmethod
    def is_path_snirf(path:str) -> bool:
        return os.path.isfile(path) and path.endswith('.snirf')

    def list_all_files(self) -> List[str]:
        """
        Recursively list all the NIRS files found in the source paths

        Returns:
            List[str]: list of absolute file paths
        """
        file_paths = []
        for root_path in self.paths:
            for path in Path(root_path).rglob('*'):
                if DataBrowser.is_path_fif(str(path)):
                    file_paths.append(str(path.absolute()))

                elif DataBrowser.is_path_nirx(str(path)):
                    file_paths.append(str(path.absolute()))

                elif DataBrowser.is_path_snirf(str(path)):
                    file_paths.append(str(path.absolute()))

        # remove duplicates
        unique = list(set(file_paths))
        unique.sort()
        return unique

    def download_demo_dataset(self) -> str:
        """
        Download a publicly available demo NIRS dataset of dyads recordings

        Source: https://researchdata.ntu.edu.sg/dataset.xhtml?persistentId=doi:10.21979/N9/35DNCW

        Returns:
            str: the local path where the dataset has been downloaded
        """
        extract_path = self.absolute_path(DOWNLOADS_RELATIVE_PATH)
        zip_path = pooch.retrieve(
            fname="fathers.zip",
            url="https://researchdata.ntu.edu.sg/api/access/datafile/91950?gbrecs=true",
            known_hash="md5:786e0c13caab4fc744b93070999dff63",
            progressbar=True
        )

        target_path = os.path.join(extract_path, 'fathers')

        if not os.path.exists(target_path):
            with ZipFile(zip_path, 'r') as zip:
                print(f'Extracting to {extract_path}, (target: {target_path})')
                zip.extractall(path=extract_path)

        self.add_source(target_path)
        return target_path
    



from typing import Tuple, List
import re
import warnings

import numpy as np
import pandas as pd

from ..wavelet.base_wavelet import BaseWavelet
from ..wavelet.implementations.pywavelets_wavelet import ComplexMorletWavelet
from ..wavelet.wtc import WTC
from ..wavelet.pair_signals import PairSignals
from ..wavelet.coherence_data_frame import CoherenceDataFrame
from ..utils import TaskList, TASK_NAME_WHOLE_RECORD
from .recording import Recording
from .preprocessor.base_preprocessor import BasePreprocessor
from ..plots import (
    plot_wavelet_transform_weights,
    plot_coherence_matrix,
    plot_coherence_bars_per_task,
    plot_coherence_connectogram,
    plot_coherence_connectogram_split,
)

PairChannelMatchSingleType = str | List[str] | re.Pattern
PairChannelMatchType = PairChannelMatchSingleType | Tuple[PairChannelMatchSingleType, PairChannelMatchSingleType]

MIN_SECTION_LENGTH = 10

class Dyad:
    """
    The Dyad object is a pair of recordings (per subject) of an hyperscanning recording.
    Their recorded channels should be time aligned.

    Args:
        s1 (Recording): recording of subject 1 of the dyad
        s2 (Recording): recording of subject 2 of the dyad
        label (str, optional): Custom label for the dyad. Defaults to `s1.label`-`s2.label`.
        is_pseudo (bool, optional): If the dyad is a permutated pair created for comparison. Used to track dyad "type" in results. Defaults to False.
    """
    s1: Recording
    s2: Recording
    label: str
    is_pseudo: bool
    tasks: TaskList # intersection of tasks of subject 1 and subject 2
    wtcs: List[WTC] | None # the computed Wavelet Transform Coherence for each channel pairs in the dyad
    df: CoherenceDataFrame | None # pandas dataframe from computed coherence

    def __init__(self, s1:Recording, s2:Recording, label:str='', is_pseudo:bool=False):
        self.s1 = s1
        self.s2 = s2
        self.wtcs = None
        self.df = None
        self.is_pseudo = is_pseudo
        self.is_intra = s1 == s2

        self.label = label
        if self.label == '':
            self.label = Dyad._get_label_from_recordings(s1, s2)

        # Intersect the tasks
        self.tasks = []
        s1_tasks = s1.tasks
        s2_tasks = s2.tasks

        s2_tasks_names = [t.name for t in s2_tasks]
        found_tasks_names = []
        for task in s1_tasks:
            task_name = task.name
            if task_name in s2_tasks_names:
                self.tasks.append(task)
                found_tasks_names.append(task_name)
    
    @property 
    def recordings(self) -> Tuple[Recording, Recording]:
        return (self.s1, self.s2)
    
    @property
    def is_preprocessed(self) -> bool:
        for recording in self.recordings:
            if not recording.is_preprocessed:
                return False
        return True
    
    @property
    def is_wtc_computed(self):
        return self.wtcs is not None

    @staticmethod
    def _get_label_from_recordings(s1:Recording, s2:Recording) -> str:
        return f'{s1.subject_label}-{s2.subject_label}'

    def preprocess(self, preprocessor: BasePreprocessor):
        """
        Run the preprocess pipeline on every subject recordings in the dyad

        Args:
            preprocessor (BasePreprocessor): Which preprocessor class to use. If no preprocessing is necessary, use MnePreprocessorUpstream()

        Returns:
            self: the Dyad object itself. Useful for chaining operations
        """
        for recording in self.recordings:
            recording.preprocess(preprocessor)
        return self

    def _append_pairs(self,
                      label_dyad:str,
                      s1_ch_names:List[str],
                      s2_ch_names:List[str],
                      s1_task_data:np.ndarray,
                      s2_task_data:np.ndarray,
                      s1:Recording,
                      s2:Recording,
                      task_name:str,
                      epoch_id:int,
                      is_intra_of:int|None,
                      is_pseudo:bool,
                      pairs:List[PairSignals]):
        n = s1_task_data.shape[1]
        x = np.linspace(0, n/s1.preprocessed.info['sfreq'], n)
        for s1_i, s1_ch_name in enumerate(s1_ch_names):
            for s2_i, s2_ch_name in enumerate(s2_ch_names):
                y1 = s1_task_data[s1_i,:]
                y2 = s2_task_data[s2_i,:]
                # Crop signals in case they are not the same length
                stop = min(len(y1), len(y2))
                y1 = y1[:stop] 
                y2 = y2[:stop] 

                # Look for NaN, and split in section
                section_id = 0
                nan_mask = np.isnan(y1) | np.isnan(y2)
                has_nan = np.any(nan_mask)

                if has_nan:
                    nan_idx = np.where(nan_mask)[0]+1
                    # We have to drop the first item of every split but the first split, because it contains the NaN
                    x_sections = [section for section in np.split(x, nan_idx) if len(section)>MIN_SECTION_LENGTH]
                    y1_sections = [section for section in np.split(y1, nan_idx) if len(section)>MIN_SECTION_LENGTH]
                    y2_sections = [section for section in np.split(y2, nan_idx) if len(section)>MIN_SECTION_LENGTH]
                    
                    # remove the lurking NaN in edges of our arrays
                    for i in range(len(x_sections)):
                        delete_flags = np.isnan(y1_sections[i]) | np.isnan(y2_sections[i])
                        x_sections[i] = np.delete(x_sections[i], delete_flags)
                        y1_sections[i] = np.delete(y1_sections[i], delete_flags)
                        y2_sections[i] = np.delete(y2_sections[i], delete_flags)
                else:
                    x_sections = [x]
                    y1_sections = [y1]
                    y2_sections = [y2]

                for section_id in range(len(x_sections)):
                    pairs.append(PairSignals(
                        x_sections[section_id],
                        y1_sections[section_id],
                        y2_sections[section_id],
                        label_ch1=s1_ch_name,
                        label_ch2=s2_ch_name,
                        label_roi1=s1.get_roi_from_channel(s1_ch_name),
                        label_roi2=s2.get_roi_from_channel(s2_ch_name),
                        label_s1=s1.subject_label,
                        label_s2=s2.subject_label,
                        label_dyad=label_dyad,
                        label_task=task_name,
                        epoch_idx=epoch_id,
                        section_idx=section_id,
                        is_intra=(is_intra_of is not None),
                        is_intra_of=is_intra_of,
                        is_pseudo=is_pseudo,
                    ))
            
    
    def get_pairs(self, s1:Recording, s2:Recording, label_dyad:str|None=None, ch_match:PairChannelMatchType|None=None, is_intra_of:int|None=None, is_pseudo:bool=False) -> List[PairSignals]:
        """
        Generate all the signal pairs between the 2 subjects and returns them in a format suitable for signal processing

        Args:
            s1 (Recording): recording of subject 1 of the dyad
            s2 (Recording): recording of subject 2 of the dyad
            label_dyad (str | None, optional): custom label for the dyad. Defaults to self.label.
            ch_match (PairMatchType, optional): string, list or regex to match channel name.
                                                Can be a tuple of 2 items if subject1 and subject2 have different matches.
                                                Defaults to None, which means all channels.
            is_pseudo (bool, optional): True if the pair is a permutated pair. Defaults to False.

        Returns:
            List[PairSignals]: a list of all the possible pairs of s1 channels with s2 channels
        """
        if label_dyad is None:
            label_dyad = self.label

        pairs: List[PairSignals] = []

        if s1.preprocessed.info['sfreq'] != s2.preprocessed.info['sfreq']:
            raise RuntimeError('Recordings must have the same sampling frequency')

        # Force match in tuple for leaner code below
        if not isinstance(ch_match, tuple):
            ch_match = (ch_match, ch_match)

        def check_match(ch_name, m):
            if m is None:
                return True
            if isinstance(m, re.Pattern):
                return m.search(ch_name) is not None
            if isinstance(m, List):
                return ch_name in m
            return m in ch_name


        s1_ch_names = [ch_name for ch_name in s1.ordered_ch_names if check_match(ch_name, ch_match[0])]
        s2_ch_names = [ch_name for ch_name in s2.ordered_ch_names if check_match(ch_name, ch_match[1])]

        seen_tasks = set()
        for task in self.tasks:
            if task.name in seen_tasks:
                continue
            seen_tasks.add(task.name)
            if task.name == TASK_NAME_WHOLE_RECORD:
                s1_task_data = s1.preprocessed.copy().pick(s1_ch_names).get_data()
                s2_task_data = s2.preprocessed.copy().pick(s2_ch_names).get_data()
                epoch_id = 0
                self._append_pairs(
                    label_dyad,
                    s1_ch_names,
                    s2_ch_names,
                    s1_task_data,
                    s2_task_data,
                    s1,
                    s2,
                    task.name,
                    epoch_id,
                    is_intra_of,
                    is_pseudo,
                    pairs,
                )
            else:
                epochs1 = s1.get_epochs_for_task(task.name).copy().pick(s1_ch_names)
                epochs2 = s2.get_epochs_for_task(task.name).copy().pick(s2_ch_names)
                if len(epochs1) != len(epochs2):
                    warnings.warn("The 2 recordings do not have the same epochs count. Some epochs will be skipped in pairs.")
                
                # add one pair for each epoch
                for i in range(min(len(epochs1), len(epochs2))):
                    s1_task_data = epochs1.get_data(copy=False)[i,:,:]
                    s2_task_data = epochs2.get_data(copy=False)[i,:,:]
                    epoch_id = i
                    self._append_pairs(
                        label_dyad,
                        s1_ch_names,
                        s2_ch_names,
                        s1_task_data,
                        s2_task_data,
                        s1,
                        s2,
                        task.name,
                        epoch_id,
                        is_intra_of,
                        is_pseudo,
                        pairs,
                    )

        return pairs
    
    def compute_wtcs(
        self,
        wavelet: BaseWavelet | None = None,
        ch_match: PairChannelMatchType | None = None,
        only_time_range: Tuple[float,float] | None = None,
        bin_seconds: float | None = None,
        period_cuts: List[float] | None = None,
        frequency_cuts: List[float] | None = None,
        verbose: bool = False,
        with_intra: bool = True,
        downsample: int | None = None,
        keep_wtcs: bool = True,
    ):
        """
        Compute the Wavelet Transform Coherence for every channel pairs on the dyad

        Args:
            wavelet (BaseWavelet | None): the wavelet to use for the wavelet transform. Defaults to ComplexMorletWavelet() without arguments.
            ch_match (PairMatchType | None, optional): string, list or regex to match channel name.
                                                Can be a tuple of 2 items if subject1 and subject2 have different matches.
                                                Defaults to None, which means all channels.
            only_time_range (Tuple[float,float] | None, optional): compute only a portion of the signal, defined as time range tuple (start, stop). Defaults to None.
            bin_seconds (float | None, optional): split the resulting WTC in time bins for balancing weights. Defaults to None.
            period_cuts (List[float] | None, optional): split the resulting WTC in period/frequency bins for balancing weights and finer analysis. Use this OR frequency_cuts, not both. Defaults to None.
            frequency_cuts (List[float] | None, optional): split the resulting WTC in period/frequency bins for balancing weights and finer analysis. Use this OR period_cuts, not both. Defaults to None.
            verbose (bool, optional): verbose flag. Defaults to False.
            with_intra (bool, optional): compute intra-subject as well. Defaults to False.
            downsample (int | None, optional): downsample in time the resulting WTC. Useful to save memory space and faster display. Defaults to None.
            keep_wtcs (bool, optional): if False, all the WTCs will be removed from object after the coherence dataframe has been computed. Useful to save memory space. Defaults to True.

        Returns:
            self: the Dyad object itself. Useful for chaining operations
        """
        if wavelet is None:
            wavelet = ComplexMorletWavelet()

        if period_cuts is not None and frequency_cuts is not None:
            raise RuntimeError('Cannot specify both period_cuts and frequency_cuts')

        self.wtcs = []

        pairs = self.get_pairs(self.s1, self.s2, ch_match=ch_match, is_pseudo=self.is_pseudo)

        if self.is_intra:
            # Force with_intra when subject 1 is the same as subject 2
            with_intra = True
        else:
            for pair in pairs:
                if verbose:
                    print(f'Running Wavelet Coherence for dyad "{self.label}" on pair "{pair.label}"')
                if only_time_range is not None:
                    pair = pair.sub(only_time_range)
                wtc = wavelet.wtc(pair, bin_seconds=bin_seconds, period_cuts=period_cuts, frequency_cuts=frequency_cuts)
                if downsample is not None:
                    wtc.downsample_in_time(downsample)

                self.wtcs.append(wtc)

        if with_intra:
            if self.is_intra:
                subjects = [self.s1]
            else:
                subjects = [self.s1, self.s2]

            for i, recording in enumerate(subjects):
                recording.intra_wtcs = []
                # if we have different channels for each recording, the intra wtc should use only the ones for this subject
                if isinstance(ch_match, tuple):
                    ch_match_intra = ch_match[i]
                else:
                    ch_match_intra = ch_match
                    
                # We have to keep track of the subject identifier when we are intra, since they would be displayed differently (parent VS child for example)
                is_intra_of = i+1
                for pair in self.get_pairs(recording, recording, f'{recording.subject_label}(intra)', ch_match=ch_match_intra, is_intra_of=is_intra_of):
                    if verbose:
                        print(f'Running Wavelet Coherence intra-subject "{recording.subject_label}" on pair "{pair.label}"')
                    if only_time_range is not None:
                        pair = pair.sub(only_time_range)
                    wtc = wavelet.wtc(pair, bin_seconds=bin_seconds, period_cuts=period_cuts, frequency_cuts=frequency_cuts)
                    if downsample is not None:
                        wtc.downsample_in_time(downsample)
                    recording.intra_wtcs.append(wtc)

        self.df = self._get_coherence_df(with_intra=with_intra)

        if not keep_wtcs:
            self.wtcs = []
            self.s1.intra_wtcs = []
            self.s1.intra_wtcs = []

        return self
    
    def _get_coherence_df(self, with_intra=False) -> pd.DataFrame:
        if with_intra:
            if not self.s1.is_wtc_computed or not self.s2.is_wtc_computed:
                raise RuntimeError('Intra subject WTCs are not computed. Please check "compute_wtcs" arguments')
            wtcs = self.wtcs + self.s1.intra_wtcs + self.s2.intra_wtcs
        else:
            wtcs = self.wtcs

        frame_rows = []
        for wtc in wtcs:
            frame_rows = frame_rows + wtc.as_frame_rows

        return CoherenceDataFrame.from_wtc_frame_rows(frame_rows)

    def reset(self):
        self.s1.intra_wtcs = None
        self.s2.intra_wtcs = None
        self.wtcs = None
        self.df = None
        return self

    #
    # Plots
    # 
    def plot_wtc(self, wtc: WTC, **kwargs):
        """
        Plot the Wavelet Transform Coherence

        Args:
            wtc (WTC): WTC object
        """
        return plot_wavelet_transform_weights(
            wtc.W,
            wtc.times,
            wtc.frequencies,
            wtc.coif,
            wtc.sfreq,
            title=wtc.label_pair,
            **kwargs)

    def plot_coherence_matrix(self, field1:str='channel1', field2:str='channel2', query:str | None=None, **kwargs):
        """
        Plot the computed coherence metric for pair of fields (channel or roi) in a matrix format

        Args:
            field1 (str): name of the field in dataframe for x axis. Defaults to 'channel1'.
            field2 (str): name of the field in dataframe for y axis. Defaults to 'channel2'.
            query (str | None, optional): pandas query to filter the dataframe. Defaults to None.
        """
        df = self.df
        if query is not None:
            df = df.query(query)
            
        ch_names1 = self.s1.ordered_ch_names
        ch_names2 = self.s2.ordered_ch_names
        ordered_names = ch_names1 + [name for name in ch_names2 if name not in ch_names1] 

        return plot_coherence_matrix(
            df,
            self.s1.subject_label,
            self.s2.subject_label,
            field1,
            field2,
            ordered_names,
            **kwargs)
        
    def plot_coherence_matrix_per_channel(self, query:str|None=None, **kwargs):
        """
        Wraps plot_coherence_matrix to plot per channel

        Args:
            query (str | None, optional): pandas query to filter the dataframe. Defaults to None.
        """
        return self.plot_coherence_matrix(
            'channel1',
            'channel2',
            query,
            **kwargs)
        
    def plot_coherence_matrix_per_roi(self, query:str|None=None, **kwargs):
        """
        Wraps plot_coherence_matrix to plot per region of interest

        Args:
            query (str | None, optional): pandas query to filter the dataframe. Defaults to None.
        """
        return self.plot_coherence_matrix(
            'roi1',
            'roi2',
            query,
            **kwargs)
    
    def plot_coherence_matrix_per_channel_for_task(self, task:str, **kwargs):
        """
        Wraps plot_coherence_matrix_per_channel to plot for a specific task

        Args:
            task (str): task name
        """
        return self.plot_coherence_matrix(
            'channel1',
            'channel2',
            query=f'task=="{task}"',
            **kwargs)
        
    def plot_coherence_matrix_per_roi_for_task(self, task:str, **kwargs):
        """
        Wraps plot_coherence_matrix_per_roi to plot for a specific task

        Args:
            task (str): task name
        """
        return self.plot_coherence_matrix(
            'roi1',
            'roi2',
            query=f'task=="{task}"',
            **kwargs)
    
    def plot_coherence_bars_per_task(self, **kwargs):
        """
        Plot coherence metric per task for comparison
        """
        return plot_coherence_bars_per_task(
            self.df,
            **kwargs)
        
    #
    # Plot connectogram (Proof of Concept)
    # 
    def plot_coherence_connectogram_intra(self, recording:Recording, query:str|None=None, **kwargs):
        df = self.df
        selector = df['is_intra']==True
        df_filtered = df[selector]

        if query is not None:
            df_filtered = df_filtered.query(query)

        pivot = df_filtered.pivot_table(index='roi1', columns='roi2', values='coherence', aggfunc='mean')
        if 'title' not in kwargs:
            kwargs['title'] = recording.subject_label

        return plot_coherence_connectogram(
            pivot,
            **kwargs)

    def plot_coherence_connectogram_s1(self, query:str|None=None, **kwargs):
        return self.plot_coherence_connectogram_intra(
            self.s1,
            query,
            **kwargs)

    def plot_coherence_connectogram_s2(self, query:str|None=None, **kwargs):
        return self.plot_coherence_connectogram_intra(
            self.s2,
            query,
            **kwargs)

    def plot_coherence_connectogram(self, query:str|None=None, title:str|None=None, **kwargs):
        df = self.df.copy()
        selector = df['is_intra']==False
        df_filtered = df[selector]

        if query is not None:
            df_filtered = df_filtered.query(query)

        # rename to have them separated in the plot
        df_filtered.loc[:, 'roi1'] = 's1_' + df_filtered['roi1'].astype(str)
        df_filtered.loc[:, 'roi2'] = 's2_' + df_filtered['roi2'].astype(str)

        pivot = df_filtered.pivot_table(index='roi1', columns='roi2', values='coherence', aggfunc='mean')

        if title is None:
            title=f'{self.s1.subject_label} / {self.s2.subject_label}'

        return plot_coherence_connectogram_split(
            pivot,
            title=title,
            **kwargs)




from abc import ABC, abstractmethod
from typing import List, Generic, TypeVar

PREPROCESS_STEP_BASE_KEY = 'base'
PREPROCESS_STEP_BASE_DESC = 'Loaded data'

PREPROCESS_STEP_OD_KEY = 'od'
PREPROCESS_STEP_OD_DESC = 'Optical density'

PREPROCESS_STEP_OD_CLEAN_KEY = 'od_clean'
PREPROCESS_STEP_OD_CLEAN_DESC = 'Optical density cleaned'

PREPROCESS_STEP_HAEMO_KEY = 'haemo'
PREPROCESS_STEP_HAEMO_DESC = 'Hemoglobin'

PREPROCESS_STEP_HAEMO_FILTERED_KEY = 'haemo_filtered'
PREPROCESS_STEP_HAEMO_FILTERED_DESC = 'Hemoglobin Band-pass Filtered'

# Generic type for underlying fnirs implementation (mne raw / cedalion recording)
T = TypeVar('T')

class BaseStep(ABC, Generic[T]):
    """
    Stores the results of a step in a preprocess pipeline

    Args:
        obj (T): the implementation dependent object representing the step processed data
        key (str): identifier for the step
        desc (str | None, optional): description of the setup. Defaults to "key" value.
    """
    obj: T
    key: str
    desc: str

    def __init__(self, obj:T, key:str, desc:str|None=None):
        self.obj = obj
        self.key = key
        if desc is None:
            self.desc = key
        else:
            self.desc = desc

    @property
    @abstractmethod
    def n_times(self) -> int:
        pass

    @property
    @abstractmethod
    def sfreq(self) -> float:
        pass

    @property
    @abstractmethod
    def ch_names(self) -> List[str]:
        pass

    @property
    def duration(self) -> float:
        return self.n_times / self.sfreq

    @abstractmethod
    def plot(self, **kwargs):
        pass



from abc import ABC, abstractmethod
from typing import List, Generic, TypeVar

from .base_step import BaseStep

# Generic type for underlying fnirs implementation (mne raw / cedalion recording)
T = TypeVar('T')

class BasePreprocessor(ABC, Generic[T]):
    @abstractmethod
    def read_file(self, path:str, verbose:bool=False) -> T:
        """
        Load the file to be preprocessed. Does not run the preprocess yet

        Args:
            path (str): NIRS file location
            verbose (bool, optional): verbosity flag. Defaults to False.

        Returns:
            T: returns an implementation dependant object representation of the loaded data
        """
        pass

    @abstractmethod
    def run(self, raw:T, verbose:bool=False):
        """
        run all the preprocessing steps on a raw object

        Args:
            raw (T): implementation dependant object representation of the loaded data. Use with "read_file()"
            verbose (bool, optional): verbosity flag. Defaults to False.

        Returns:
            List[BasePreprocessStep[T]]: list of all the intermediary steps of the preprocessing
        """
        pass




from typing import List
from collections import OrderedDict

import scipy.io

class ChannelROI:
    rois: OrderedDict

    def __init__(self, rois:OrderedDict|None=None):
        """
        Structure for grouping channels into regions of interest

        Args:
            rois (OrderedDict | None, optional): predefined regions of interest. Defaults to None.
        """
        if rois is None:
            self.rois = OrderedDict()
        else:
            self.rois = rois

    @staticmethod
    def from_lionirs_file(roi_file_path:str):
        """
        Create a new ChannelROI object from a file exported 

        Args:
            roi_file_path (str): The path of the file exported by LIONirs toolbox for Region of Interests
                                 It should be a matlab file with structure zone->(label|plotLst)

        Returns:
            ChannelROI: an instance of the class
        """
        obj = ChannelROI()

        mat = scipy.io.loadmat(roi_file_path, struct_as_record=False)

        row_channel_name_map = dict()

        for i, channel_row in enumerate(mat['zone'][0][0].ml):
            row_id = i+1
            source_id = channel_row[0]
            detector_id = channel_row[1]
            # The last column is 1 or 2, which represents hbo or hbr, but we don't need that info here
            row_channel_name_map[row_id] = f'S{source_id}_D{detector_id}'

        mat_rois = mat['zone'][0][0].label[0]
        mat_rows_per_rois = mat['zone'][0][0].plotLst[0]

        for mat_roi, mat_rows_per_roi in zip(mat_rois, mat_rows_per_rois):
            roi_name = mat_roi[0]
            rows_in_roi = mat_rows_per_roi[0]
            channels_in_roi = []
            for row in rows_in_roi:
                channels_in_roi.append(row_channel_name_map[row])
            obj.rois[roi_name] = channels_in_roi
        return obj
        
    @property
    def ordered_ch_names(self):
        channel_names = []
        for channel_names_in_roi in self.rois.values():
            channel_names = channel_names + channel_names_in_roi
        return channel_names
    
    @property
    def group_boundaries_sizes(self):
        boundaries = [0]
        for values in self.rois.values():
            boundaries.append(len(values) + boundaries[-1])
        boundaries.append(len(self.ordered_ch_names))
        return boundaries

    def get_ch_names_in_order(self, ch_names:str) -> List[str]:
        """
        Given a list of channel names, return them in order of Region of Interest

        Args:
            names (str): 

        Returns:
            List[str]: all the received ch_names, ordered according to defined regions of interest
        """
        all_names_found = []

        # OPTIMIZE: this is inefficient
        for ordered_name in self.ordered_ch_names:
            names_found = [name for name in ch_names if name.startswith(ordered_name)]
            for name_found in names_found:
                if name_found not in all_names_found:
                    all_names_found.append(name_found)
        
        for unordered_ch_name in ch_names:
            if not unordered_ch_name in all_names_found:
                all_names_found.append(unordered_ch_name)
        
        return all_names_found

    def get_roi_from_channel(self, ch_name):
        for k, v in self.rois.items():
            # OPTIMIZE: this is inefficient
            names_found = [name for name in v if ch_name.startswith(name)]
            if len(names_found) > 0:
                return k
        return ''
        




from typing import List, Tuple
import pickle

from ..wavelet.base_wavelet import BaseWavelet
from .preprocessor.base_preprocessor import BasePreprocessor
from ..wavelet.coherence_data_frame import CoherenceDataFrame
from ..profiling import TimeTracker
from .dyad import Dyad, PairChannelMatchType
from ..plots import (
    plot_coherence_matrix,
    plot_coherence_bars_per_task,
    plot_coherence_connectogram,
    plot_coherence_connectogram_split,
)

class Study():
    dyads: List[Dyad]
    dyads_shuffled: List[Dyad]|None

    def __init__(self, dyads: List[Dyad] = []):
        """
        The Study object encapsulates the logic of an hyperscanning experiment.

        Args:
            dyads (List[Dyad], optional): List of all the dyads in the study. Defaults to [].
        """
        self.dyads = dyads
        self.dyads_shuffled = None

    @property
    def is_wtc_computed(self) -> bool:
        """True if the WTCs have been computed on all dyads in the study object"""
        for dyad in self.dyads:
            if not dyad.is_wtc_computed:
                return False
        return True

    @property
    def is_wtc_shuffle_computed(self) -> bool:
        """True if the WTCs have been computed on all shuffled dyads in the study object"""
        if self.dyads_shuffled is None:
            return False
        for dyad in self.dyads_shuffled:
            if not dyad.is_wtc_computed:
                return False
        return True
    
    @property
    def df(self) -> CoherenceDataFrame:
        """The pandas dataframe object from computed WTCs"""
        return self.get_coherence_df()

    def preprocess(self, preprocessor: BasePreprocessor):
        """
        Run the preprocess steps on every recordings in the study

        Args:
            preprocessor (BasePreprocessor): Which preprocessor to use. If no preprocessing is necessary, use MnePreprocessorUpstream()

        Returns:
            self: the Study object itself. Useful for chaining operations
        """
        for dyad in self.dyads:
            if not dyad.is_preprocessed:
                dyad.preprocess(preprocessor)
        return self
    
    def compute_wtcs(
        self,
        # All the arguments are a copy from Dyad.compute_wtcs, but we need them explicitely for type hinting
        wavelet: BaseWavelet|None = None,
        ch_match: PairChannelMatchType|None = None,
        only_time_range: Tuple[float,float]|None = None,
        bin_seconds: float|None = None,
        period_cuts: List[float]|None = None,
        frequency_cuts: List[float]|None = None,
        verbose: bool = False,
        with_intra: bool = True,
        downsample: int|None = None,
        keep_wtcs: bool = True,
        show_time_estimation: bool = True,
    ): 
        """
        Wraps the `compute_wtcs` of all the dyads. Arguments are directly passed to the dyads method

        Args:
            wavelet (BaseWavelet | None): the wavelet to use for the wavelet transform. Defaults to ComplexMorletWavelet() without arguments.
            ch_match (PairMatchType | None, optional): string, list or regex to match channel name.
                                                Can be a tuple of 2 items if subject1 and subject2 have different matches.
                                                Defaults to None, which means all channels.
            only_time_range (Tuple[float,float] | None, optional): compute only a portion of the signal, defined as time range. Defaults to None.
            bin_seconds (float | None, optional): split the resulting WTC in time bins for balancing weights. Defaults to None.
            period_cuts (List[float] | None, optional): split the resulting WTC in period/frequency bins for balancing weights and finer analysis. Defaults to None.
            frequency_cuts (List[float] | None, optional): split the resulting WTC in period/frequency bins for balancing weights and finer analysis. Defaults to None.
            verbose (bool, optional): verbose flag. Defaults to False.
            with_intra (bool, optional): compute intra-subject as well. Defaults to False.
            downsample (int | None, optional): downsample in time the resulting WTC. Useful to save memory space and faster display. Defaults to None.
            keep_wtcs (bool, optional): if False, all the WTCs will be removed from object after the coherence dataframe has been computed. Useful to save memory space. Defaults to True.

        Returns:
            self: the Study object itself. Useful for chaining operations
        """
        for i, dyad in enumerate(self.dyads):
            if i == 0:
                tracker = TimeTracker()
                tracker.start()

            dyad.compute_wtcs(
                wavelet,
                ch_match,
                only_time_range,
                bin_seconds,
                period_cuts,
                frequency_cuts,
                verbose,
                with_intra,
                downsample,
                keep_wtcs,
            )

            if i == 0:
                tracker.stop()
                if show_time_estimation:
                    self._print_time_estimation(tracker.duration, len(self.dyads))
                
        return self
    
    def _print_time_estimation(self, single_duration, count):
            print(f'Time for computing one dyad: {TimeTracker.human_readable_duration(single_duration)}')
            print(f'Expected time for {count} dyads: {TimeTracker.human_readable_duration(single_duration * count)}')
    
    def estimate_wtcs_run_time(self, *args, **kwargs):
        """
        Computes the WTC for one dyad and print the expected run time for the whole study

        Returns:
            self: the Study object itself. Useful for chaining operations
        """
        dyad = self.dyads[0]
        tracker = TimeTracker()
        tracker.start()
        dyad.compute_wtcs(*args, **kwargs)
        tracker.stop()
        self._print_time_estimation(tracker.duration, len(self.dyads))

        return self
    
    def reset(self):
        self._clear_dyads_shuffle()
        for dyad in self.dyads:
            dyad.reset()
        return self
    
    def _clear_dyads_shuffle(self):
        """
        Delete all the shuffle dyads that have been created

        Returns:
            self: the Study object itself. Useful for chaining operations
        """
        self.dyads_shuffled = None
        return self
    
    def get_dyads_shuffle(self) -> List[Dyad]:
        """
        Get a list of permutated recording pairs, useful for statistical analysis.

        Returns:
            List[Dyad]: permulated pairs
        """
        dyads_shuffle = []
        for i, dyad1 in enumerate(self.dyads):
            for j, dyad2 in enumerate(self.dyads):
                if i == j:
                    continue
                dyads_shuffle.append(Dyad(dyad1.s1, dyad2.s2, label=f'shuffle s1:{dyad1.label}-s2:{dyad2.label}', is_pseudo=True))
        return dyads_shuffle

    # TODO add as argument the number of shuffle dyads
    def compute_wtcs_shuffle(self, *args, **kwargs):
        """
        Wraps the `compute_wtcs` of all the dyads_shuffle. Arguments are directly passed to the dyads_shuffle method

        Returns:
            self: the Study object itself. Useful for chaining operations
        """
        self.dyads_shuffled = self.get_dyads_shuffle()
        for dyad_shuffle in self.dyads_shuffled:
            dyad_shuffle.compute_wtcs(*args, **kwargs, with_intra=False)
        return self
    
    def get_coherence_df(self) -> CoherenceDataFrame:
        """
        Loop over every dyad and concatenate all the pandas dataframes into one

        If dyads_shuffle are computed, they are also included

        Raises:
            RuntimeError: all the WTCs must have been computed in order to get the pandas dataframe

        Returns:
            CoherenceDataFrame: typed pandas dataframe object
        """
        dfs = []
        if not self.is_wtc_computed:
            raise RuntimeError('wtc not computed')

        for dyad in self.dyads:
            dfs.append(dyad.df)

        if self.is_wtc_shuffle_computed:
            for dyad_shuffle in self.dyads_shuffled:
                dfs.append(dyad_shuffle.df)

        return CoherenceDataFrame.concat(dfs)
    
    #
    # Plots
    #
    def plot_coherence_matrix(
            self,
            field1:str='channel1',
            field2:str='channel2',
            query:str | None=None,
            s1_label:str='Subject1',
            s2_label:str='Subject2',
            **kwargs):
        """
        Plot the computed coherence metric for pair of fields (channel or roi) in a matrix format

        Args:
            field1 (str): name of the field in dataframe for x axis. Defaults to 'channel1'.
            field2 (str): name of the field in dataframe for y axis. Defaults to 'channel2'.
            query (str | None, optional): pandas query to filter the dataframe. Defaults to None.
        """
        df = self.df
        if query is not None:
            df = df.query(query)
            
        ch_names1 = []
        ch_names2 = []

        for dyad in self.dyads:
            for ch_name in dyad.s1.ordered_ch_names:
                if ch_name not in ch_names1:
                    ch_names1.append(ch_name)

            for ch_name in dyad.s2.ordered_ch_names:
                if ch_name not in ch_names2:
                    ch_names2.append(ch_name)

        ordered_names = ch_names1 + [name for name in ch_names2 if name not in ch_names1] 
        # TODO: missing ordered roi

        return plot_coherence_matrix(
            df,
            s1_label,
            s2_label,
            field1,
            field2,
            ordered_names,
            **kwargs)
        
    def plot_coherence_matrix_per_channel(self, query:str|None=None, **kwargs):
        """
        Wraps plot_coherence_matrix to plot per channel

        Args:
            query (str | None, optional): pandas query to filter the dataframe. Defaults to None.
        """
        return self.plot_coherence_matrix(
            'channel1',
            'channel2',
            query,
            **kwargs)
        
    def plot_coherence_matrix_per_roi(self, query:str|None=None, **kwargs):
        """
        Wraps plot_coherence_matrix to plot per region of interest

        Args:
            query (str | None, optional): pandas query to filter the dataframe. Defaults to None.
        """
        return self.plot_coherence_matrix(
            'roi1',
            'roi2',
            query,
            **kwargs)
    
    def plot_coherence_matrix_per_channel_for_task(self, task:str, **kwargs):
        """
        Wraps plot_coherence_matrix_per_channel to plot for a specific task

        Args:
            task (str): task name
        """
        return self.plot_coherence_matrix(
            'channel1',
            'channel2',
            query=f'task=="{task}"',
            **kwargs)
        
    def plot_coherence_matrix_per_roi_for_task(self, task:str, **kwargs):
        """
        Wraps plot_coherence_matrix_per_roi to plot for a specific task

        Args:
            task (str): task name
        """
        return self.plot_coherence_matrix(
            'roi1',
            'roi2',
            query=f'task=="{task}"',
            **kwargs)
    
    def plot_coherence_bars_per_task(self, **kwargs):
        """
        Plot coherence metric per task for comparison
        """
        return plot_coherence_bars_per_task(
            self.df,
            **kwargs)
        
    def plot_coherence_connectogram(
            self,
            query:str|None=None,
            title:str|None=None,
            s1_label:str='Subject1',
            s2_label:str='Subject2',
            **kwargs):
        df = self.df.copy()
        selector = df['is_intra']==False
        df_filtered = df[selector]

        if query is not None:
            df_filtered = df_filtered.query(query)

        # rename to have them separated in the plot
        df_filtered.loc[:, 'roi1'] = s1_label + '_' + df_filtered['roi1'].astype(str)
        df_filtered.loc[:, 'roi2'] = s2_label + '_' + df_filtered['roi2'].astype(str)

        pivot = df_filtered.pivot_table(index='roi1', columns='roi2', values='coherence', aggfunc='mean')

        if title is None:
            title = f'{s1_label} / {s2_label}'

        return plot_coherence_connectogram_split(
            pivot,
            title=title,
            **kwargs)

    def plot_coherence_connectogram_intra(self, is_intra_of:int, query:str|None=None, **kwargs):
        df = self.df
        selector = (df['is_intra']==True) & (df['is_intra_of']==is_intra_of)
        df_filtered = df[selector]

        if query is not None:
            df_filtered = df_filtered.query(query)

        pivot = df_filtered.pivot_table(index='roi1', columns='roi2', values='coherence', aggfunc='mean')
        return plot_coherence_connectogram(pivot, **kwargs)

    def plot_coherence_connectogram_s1(self, query:str|None=None, **kwargs):
        return self.plot_coherence_connectogram_intra(1, query, **kwargs)

    def plot_coherence_connectogram_s2(self, query:str|None=None, **kwargs):
        return self.plot_coherence_connectogram_intra(2, query, **kwargs)

    #
    # Disk serialisation
    #
    @staticmethod
    def from_pickle(file_path: str):
        """
        Reload a Study object from a serialized file

        Args:
            file_path (str): previously stored study object

        Returns:
            Study: the Study object
        """
        with open(file_path, 'rb') as f:
            return pickle.load(f)

    def save_pickle(self, file_path: str):
        """
        Serialize the object to a local file

        Args:
            file_path (str): disk path for the serialisation
        """
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)

    def save_feather(self, file_path: str):
        """
        Save the pandas dataframe of coherence data to a .feather file

        Args:
            file_path (str): disk path for the .feather file
        """
        CoherenceDataFrame.save_feather(self.df, file_path)

    def save_csv(self, file_path: str):
        """
        Save the pandas dataframe of coherence data to a .csv file

        Args:
            file_path (str): disk path for the .csv file
        """
        self.df.to_csv(file_path)



import os
from pathlib import Path
import sys
import tempfile

from shiny import App, Inputs, Outputs, Session, reactive, render, ui
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import pandas as pd

root = os.path.join(Path(__file__).parent, '..', '..')
sys.path.append(root)
import hypyp.plots

from hypyp.fnirs import Study, Dyad
from hypyp.wavelet.base_wavelet import WTC

HARDCODED_RESULTS_PATH = "./data/results"

SIDEBAR_WIDTH = 400 # px
DEFAULT_PLOT_COHERENCE_MATRIX_HEIGHT = 1000
DEFAULT_PLOT_COHERENCE_PER_TASK_HEIGHT = 600
DEFAULT_PLOT_WTC_HEIGHT = 600
DEFAULT_PLOT_CONNECTOGRAM_HEIGHT = 1000

STR_ALL_DYADS = 'All dyads'
STR_ALL_TASKS = 'All tasks'

# This is to avoid having external windows launched
matplotlib.use('Agg')

def mne_figure_as_image(fig):
    temp_img_path = tempfile.NamedTemporaryFile(suffix=".png").name
    fig.savefig(temp_img_path)
    return {"src": temp_img_path, "alt": "MNE Plot"}
        
def ui_option_row(label, ui_element, sizes=(6, 6), center=False):
    label_style = ""
    if center:
        label_style += "text-align: right;"
    return ui.row(
        ui.column(sizes[0], label, style=label_style),
        ui.column(sizes[1], ui_element),
    )

# UI
app_ui = ui.page_fluid(
    ui.page_navbar(
        ui.nav_spacer(),
        ui.nav_panel(
            "Study Info",
            ui.row(
                ui.column(
                    12,
                    ui.output_table('table_info_study')
                )
            ),
        ),
        ui.nav_panel(
            "Coherence Matrix",
            ui.row(
                ui.column(
                    10,
                    ui.output_plot('plot_coherence_matrix', height=DEFAULT_PLOT_COHERENCE_MATRIX_HEIGHT)
                ),
                ui.column(
                    2,
                    ui.input_select(
                        'coherence_select_grouping',
                        'Coherence grouping',
                        choices={
                            'roi': 'Region of Interest',
                            'channel': 'Individual Channels',
                            'channel_roi': 'Channel-ROI',
                            'roi_channel': 'ROI-Channel',
                        })
                ),
            ),
        ),
        ui.nav_panel(
            "Connectograms",
            ui.row(
                ui.column(
                    4,
                    ui.output_plot('plot_connectogram_s1', height=DEFAULT_PLOT_CONNECTOGRAM_HEIGHT)
                ),
                ui.column(
                    4,
                    ui.output_plot('plot_connectogram_s2', height=DEFAULT_PLOT_CONNECTOGRAM_HEIGHT)
                ),
                ui.column(
                    4,
                    ui.output_plot('plot_connectogram', height=DEFAULT_PLOT_CONNECTOGRAM_HEIGHT)
                ),
            ),
        ),
        ui.nav_panel(
            "Coherence Per Task",
            ui.row(
                ui.column(
                    12,
                    ui.output_plot('plot_coherence_per_task', height=DEFAULT_PLOT_COHERENCE_PER_TASK_HEIGHT)
                ),
            ),
        ),
        ui.nav_panel(
            "Data Frame",
            ui.row(
                ui.column(
                    12,
                    ui.output_data_frame('data_frame')
                ),
            ),
        ),
        ui.nav_panel(
            "Wavelet Transform Coherence",
            ui.row(
                ui.column(
                    10,
                    ui.output_plot('plot_wtc', height=DEFAULT_PLOT_WTC_HEIGHT)
                ),
                ui.column(
                    2,
                    ui.output_ui('ui_input_select_wtc'),
                ),
            ),
        ),
        ui.nav_spacer(),
        selected='Study Info',
        #selected='Coherence Matrix',
        #selected='Connectograms',
        #selected='Coherence Per Task',
        #selected='Wavelet Transform Coherence',
        id='main_nav',
        sidebar=ui.sidebar(
            ui.output_ui('ui_input_study_file'),
            ui.output_ui('ui_input_select_dyad'),
            ui.output_ui('ui_input_select_task'),
            width=SIDEBAR_WIDTH,
        ),
        title="HyPyP fNIRS results viewer",
        fillable=True,
        bg="transparent"
    )
)

def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc
    def get_study() -> Study:
        study_file_name = input.select_study_file()
        if study_file_name == '':
            return None
        study_file_path = os.path.join(HARDCODED_RESULTS_PATH, study_file_name)
        return Study.from_pickle(study_file_path)

    @reactive.calc
    def get_dyad() -> Dyad:
        study = get_study()
        if study is None:
            return None

        if input.select_dyad() is None or input.select_dyad() == STR_ALL_DYADS:
            return None
            
        dyads = [dyad for dyad in study.dyads if dyad.label == input.select_dyad()]
        if len(dyads) == 0:
            return None
        return dyads[0]

    @reactive.calc
    def get_wtc() -> WTC:
        dyad = get_dyad()
        if dyad is None:
            return None
        wtcs = [wtc for wtc in dyad.wtcs if wtc.label_pair == input.select_wtc()]
        if len(wtcs) == 0:
            return None
        return wtcs[0]

    @render.ui
    def ui_input_study_file():
        my_list = [d for d in os.listdir(HARDCODED_RESULTS_PATH) if d.endswith('.pickle')]
        my_list.sort()
        my_list = [''] + my_list
        #print(my_list)
        return ui.input_select(
            "select_study_file",
            f"Study file ({HARDCODED_RESULTS_PATH})",
            choices=my_list,
            selected="fnirs_study_lionlab.pickle",
        )
    
    @render.ui
    def ui_input_select_dyad():
        study = get_study()
        if study is None:
            return None
        return ui.input_select(
            "select_dyad",
            f"Select Dyad",
            choices=[STR_ALL_DYADS] + [dyad.label for dyad in study.dyads],
        )
    
    @render.ui
    def ui_input_select_task():
        study = get_study()
        dyad = get_dyad()
        if dyad is not None:
            tasks = dyad.tasks
        elif study is not None:
            tasks = study.dyads[0].tasks
        else:
            return None

        return ui.input_select(
            "select_task",
            f"Select Task",
            choices=[STR_ALL_TASKS] + [task.name for task in tasks],
        )
    
    @render.ui
    def ui_input_select_wtc():
        dyad = get_dyad()
        if dyad is None:
            return None
        return ui.input_select(
            "select_wtc",
            f"Select Channel Pair",
            # TODO the filtering here is sketchy
            choices=[wtc.label_pair for wtc in dyad.wtcs if wtc.label_pair.startswith(input.select_task())]
        )
    
    @render.table
    def table_info_study():
        study = get_study()
        if study is None:
            return None
        return pd.DataFrame({
            'Dyad Label': [dyad.label for dyad in study.dyads],
            'Tasks': [', '.join([task.name for task in dyad.tasks]) for dyad in study.dyads],
            'Subject 1': [dyad.s1.subject_label for dyad in study.dyads],
            'Subject 2': [dyad.s2.subject_label for dyad in study.dyads],
        })
    
    def get_query():
        task = input.select_task() 
        q = f'task == "{task}"' if task != STR_ALL_TASKS else None
        return q

    @render.plot
    def plot_coherence_matrix():
        obj = get_dyad()
        if obj is None:
            obj = get_study()
        if obj is None:
            return None

        grouping = input.coherence_select_grouping()
        if grouping == 'roi':
            return obj.plot_coherence_matrix_per_roi(query=get_query())
        elif grouping == 'channel':
            return obj.plot_coherence_matrix_per_channel(query=get_query())
        elif grouping == 'roi_channel':
            return obj.plot_coherence_matrix('roi1', 'channel2', query=get_query())
        elif grouping == 'channel_roi':
            return obj.plot_coherence_matrix('channel1', 'roi2', query=get_query())
        else:
            raise RuntimeError(f'Unknown grouping {grouping}')

    @render.plot
    def plot_connectogram_s1():
        obj = get_dyad()
        if obj is None:
            obj = get_study()
        if obj is None:
            return None
        return obj.plot_coherence_connectogram_s1(query=get_query(), title='Subject1')
        
    @render.plot
    def plot_connectogram_s2():
        obj = get_dyad()
        if obj is None:
            obj = get_study()
        if obj is None:
            return None
        return obj.plot_coherence_connectogram_s2(query=get_query(), title='Subject2')
        
    @render.plot
    def plot_connectogram():
        obj = get_dyad()
        if obj is None:
            obj = get_study()
        if obj is None:
            return None
        return obj.plot_coherence_connectogram(query=get_query())
        
    @render.data_frame
    def data_frame():
        obj = get_dyad()
        if obj is None:
            obj = get_study()
        if obj is None:
            return None
        # have a maximum so that it loads
        return render.DataGrid(obj.df[:1000], selection_mode="rows")
        
    @render.plot
    def plot_coherence_per_task():
        obj = get_dyad()
        if obj is None:
            obj = get_study()
        if obj is None:
            return None
        return obj.plot_coherence_bars_per_task()
        
    @render.plot
    def plot_wtc():
        fig, ax = plt.subplots()
        wtc = get_wtc()
        if wtc is None:
            return None
        
        wtc.plot(
            ax=ax,
            show_colorbar=True,
            show_cone_of_influence=True,
            show_nyquist=True,
        )
        return fig
    

app = App(app_ui, server)


from .preprocessor.base_preprocessor import BasePreprocessor, BaseStep
from .preprocessor.implementations.mne_preprocessor_raw_to_haemo import MnePreprocessorRawToHaemo, MneStep
from .preprocessor.implementations.mne_preprocessor_as_is import MnePreprocessorAsIs
from .data_browser import DataBrowser
from .recording import Recording
from .dyad import Dyad
from .study import Study
from .channel_roi import ChannelROI

__all__ = [
    'Recording',
    'Dyad',
    'Study',
    'DataBrowser',
    'BasePreprocessor',
    'BaseStep',
    'MnePreprocessorAsIs',
    'MnePreprocessorRawToHaemo',
    'MneStep',
    'ChannelROI',
]



import mne

from ..base_step import BaseStep

class MneStep(BaseStep[mne.io.Raw]):
    @property
    def n_times(self):
        return self.obj.n_times
        
    @property
    def sfreq(self):
        return self.obj.info['sfreq']
        
    @property
    def ch_names(self):
        return self.obj.ch_names
        
    def plot(self, **kwargs):
        return self.obj.plot(**kwargs)




import os
import tempfile

from shiny import App, Inputs, Outputs, Session, reactive, render, ui
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from scipy import fft
import mne

from hypyp.wavelet.pair_signals import PairSignals
from hypyp.fnirs.data_browser import DataBrowser
from hypyp.fnirs.recording import Recording
from hypyp.fnirs.preprocessor.implementations.mne_preprocessor_raw_to_haemo import MnePreprocessorRawToHaemo
from hypyp.fnirs.preprocessor.implementations.mne_preprocessor_as_is import MnePreprocessorAsIs
from hypyp.fnirs.preprocessor.implementations.cedalion_preprocessor import CedalionPreprocessor
from hypyp.signal import SyntheticSignal
from hypyp.wavelet.base_wavelet import BaseWavelet
from hypyp.wavelet.implementations.matlab_wavelet import MatlabWavelet
from hypyp.wavelet.implementations.pywavelets_wavelet import ComplexGaussianWavelet, ComplexMorletWavelet
from hypyp.wavelet.implementations.pycwt_wavelet import PycwtWavelet


DEFAULT_PLOT_SIGNAL_HEIGHT = 150 # px
DEFAULT_PLOT_MNE_HEIGHT = 1200 # px
DEFAULT_SIGNAL_DURATION_SELECTION = 60 # seconds

# Use this environment variable to add a lookup path for NIRS data
HYPYP_NIRS_DATA_PATH = os.getenv("HYPYP_NIRS_DATA_PATH", "")

STR_SAME_AS_SUBJECT_1 = 'Same as Subject 1'

# This is to avoid having external windows launched
matplotlib.use('Agg')

def ui_option_row(label, ui_element, sizes=(6, 6), center=False):
    label_style = ""
    if center:
        label_style += "text-align: right;"
    return ui.row(
        ui.column(sizes[0], label, style=label_style),
        ui.column(sizes[1], ui_element),
    )

# UI
app_ui = ui.page_fluid(
    ui.page_navbar(
        ui.nav_spacer(),
        ui.nav_panel(
            "Data Browser",
            ui.tags.h4(ui.output_text('text_s1_file_path')),
            ui.output_ui('ui_input_s1_mne_duration_slider'),
            ui.output_ui('ui_preprocess_steps'),
        ),
        ui.nav_panel(
            "Wavelet Coherence",
            ui.row(
                ui.column(12,
                    ui.card(
                        ui.tags.strong('Choice of signals'),
                        ui.output_ui('ui_input_signal_slider'),
                        ui.output_plot('plot_signals', height=DEFAULT_PLOT_SIGNAL_HEIGHT),
                        ui.output_plot('plot_signals_spectrum', height=DEFAULT_PLOT_SIGNAL_HEIGHT),
                    ),
                ),
            ),
            ui.row(
                ui.column(6,
                    ui.card(
                        ui.tags.strong('Wavelet'),
                        ui.output_plot('plot_mother_wavelet'),
                    ),
                ),
                ui.column(6,
                    ui.card(
                        ui.tags.strong('Wavelet Coherence'),
                        ui.output_plot('plot_wtc'),
                    ),
                ),
            ),
        ),
        ui.nav_spacer(),
        #selected='Data Browser',
        selected='Wavelet Coherence',
        id='main_nav',
        sidebar=ui.sidebar(
            ui.tags.strong('Signal parameters'),
            ui.input_select(
                "signal_type",
                "",
                choices={
                    'data_files': 'Local data files',
                    'testing': 'Synthetic signals',
                },
            ),
            ui.output_ui('ui_input_signal_choice'),
            ui.output_ui('ui_input_signal_choice_property'),
            ui.output_ui('ui_input_signal_options'),

            ui.tags.strong('Wavelet parameters'),
            ui.input_select(
                "wavelet_library",
                "",
                choices={
                    'pywavelets': 'Use pywavelets',
                    'pycwt': 'Use pycwt (based on matlab code)',
                    'matlab': 'Use Matlab engine',
                },
            ),
            ui.output_ui('ui_input_wavelet_type'),
            ui.output_ui('ui_input_wavelet_options'),
            ui.output_ui('ui_input_coherence_options'),

            ui.input_action_button("button_action_compute_wtc", label="Compute Wavelet Transform Coherence"),

            ui.tags.strong('Display parameters'),
            ui_option_row("Signal offset y", ui.input_checkbox("display_signal_offset_y", "", value=True), sizes=(8,4)),
            ui_option_row("Downsample", ui.input_checkbox("display_downsample", "", value=True), sizes=(8,4)),
            ui_option_row("Show COI", ui.input_checkbox("display_show_coi", "", value=True), sizes=(8,4)),
            ui_option_row("Show Nyquist", ui.input_checkbox("display_show_nyquist", "", value=True), sizes=(8,4)),
            open="always",
            width=400,
        ),
        title="HyPyP fNIRS explorer",
        fillable=True,
        bg="transparent"
    )
)

def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc()
    def get_signals() -> PairSignals:
        pair = None
        if input.signal_type() == 'testing':
            fs = input.signal_sampling_frequency()
            N = input.signal_n()
            x = np.linspace(0, N/fs, N)
            tmax = N / fs
            if input.signal_testing_choice() == 'sinusoid':
                y1 = SyntheticSignal(duration=tmax, n_points=N).add_sin(input.signal_sinusoid_freq1()).y
                y2 = SyntheticSignal(duration=tmax, n_points=N).add_sin(input.signal_sinusoid_freq2()).y
            elif input.signal_testing_choice() == 'sinusoid_almost_similar':
                y1 = SyntheticSignal(duration=tmax, n_points=N).add_sin(input.signal_sinusoid_almost_similar_freq1()).y
                y2 = SyntheticSignal(duration=tmax, n_points=N).add_sin(input.signal_sinusoid_almost_similar_freq1() - 0.001).y
            elif input.signal_testing_choice() == 'sinusoid_dephased':
                y1 = SyntheticSignal(duration=tmax, n_points=N).add_sin(input.signal_sinusoid_dephased_freq1()).y
                y1_fft = np.fft.fft(y1)
                y1_magnitude = np.abs(y1_fft)
                y1_phase = np.angle(y1_fft)
                y1_random_phase = np.random.uniform(-np.pi, np.pi, len(y1_phase))
                y1_randomized_fft = y1_magnitude * np.exp(1j * y1_random_phase)
                y2 = np.fft.ifft(y1_randomized_fft).real
                y2 = y2 * 1 / (np.abs(np.max(y2)-np.min(y2)) / 2)
            elif input.signal_testing_choice() == 'chirp':
                signal1 = SyntheticSignal(duration=tmax, n_points=N)
                signal1.add_chirp(input.signal_chirp_freq1(), input.signal_chirp_freq2())
                y1 = signal1.y
                y2 = np.flip(y1)
            elif input.signal_testing_choice() == 'chirp_sinusoid':
                y1 = SyntheticSignal(duration=tmax, n_points=N).add_sin(input.signal_chirp_sinusoid_freq1()).y
                y2 = SyntheticSignal(duration=tmax, n_points=N).add_chirp(input.signal_chirp_sinusoid_freq2(), input.signal_chirp_sinusoid_freq3()).y

            noise_level = input.signal_noise_level()
            y1 +=  noise_level * np.random.normal(0, 1, len(x))
            y2 +=  noise_level * np.random.normal(0, 1, len(x))

        elif input.signal_type() == 'data_files':
            # TODO this only works with MNE
            s1_raw = get_recording_s1_step().obj.copy()
            s2_raw = get_recording_s2_step().obj.copy()

            s1_selected_ch = s1_raw.pick(mne.pick_channels(s1_raw.ch_names, include = [input.signal_data_files_s1_channel()]))
            s2_selected_ch = s2_raw.pick(mne.pick_channels(s2_raw.ch_names, include = [input.signal_data_files_s2_channel()]))

            y1 = np.ma.masked_invalid(s1_selected_ch.get_data())[0,:]
            y2 = np.ma.masked_invalid(s2_selected_ch.get_data())[0,:]

            # Crop signals
            stop = min(len(y1), len(y2))
            y1 = y1[:stop] 
            y2 = y2[:stop] 
            
            # Force N for this signal
            N = len(y1)
            fs = s1_selected_ch.info['sfreq']
            x = np.linspace(0, N/fs, N)

        if pair is None:
            pair = PairSignals(
                x,
                y1,
                y2,
                label_s1='s1',
                label_s2='s2',
                label_task='foo',
            )

        try:
            if input.signal_range() is not None:
                return pair.sub(input.signal_range())
        except:
            pass
        
        return pair

    def get_signal_duration():
        if input.signal_type() == 'testing':
            fs = input.signal_sampling_frequency()
            N = input.signal_n()
            return N/fs
        if input.signal_type() == 'data_files':
            return min([get_recording_s1_step().duration, get_recording_s2_step().duration])
            

    @reactive.calc()
    def get_wavelet():
        if input.wavelet_library() == 'pywavelets':
            wavelet_name = input.wavelet_name()
            if wavelet_name == 'cmor':
                return ComplexMorletWavelet(
                    bandwidth_frequency=input.wavelet_bandwidth(),
                    center_frequency=input.wavelet_center_frequency(),
                    period_range=(input.wavelet_period_range_low(), input.wavelet_period_range_high()),
                    disable_caching=True,
                )
            elif wavelet_name == 'cgau':
                return ComplexGaussianWavelet(
                    degree=int(input.wavelet_degree()),
                    period_range=(input.wavelet_period_range_low(), input.wavelet_period_range_high()),
                    wtc_smoothing_win_size=input.smoothing_win_size(),
                    disable_caching=True,
                )
            else:
                raise RuntimeError(f'Unknown wavelet_name: {wavelet_name}')

        if input.wavelet_library() == 'pycwt':
            return PycwtWavelet()

        if input.wavelet_library() == 'matlab':
            return MatlabWavelet()

        raise RuntimeError(f'Unknown wavelet library: {input.wavelet_library()}')

    @reactive.event(input.button_action_compute_wtc)
    def compute_coherence():
        wavelet = get_wavelet()
        pair = get_signals()
        wtc = wavelet.wtc(pair)
        if input.display_downsample():
            wtc.downsample_in_time(500)
        return wtc

    def get_signal_data_files_s1_path():
        return input.signal_data_files_s1_path()
        
    def get_signal_data_files_s2_path():
        value = input.signal_data_files_s2_path()
        if value == STR_SAME_AS_SUBJECT_1:
            return input.signal_data_files_s1_path()
        return value

    @render.text
    def text_s1_file_path():
        return f"File: {get_signal_data_files_s1_path()}"

    @render.text
    def text_s2_file_path():
        value = input.signal_data_files_s2_path()
        return f"File: {get_signal_data_files_s2_path()}"

    @render.text
    def text_info_s1_file_path():
        return f"File: {get_signal_data_files_s1_path()}"

    @render.text
    def text_info_s2_file_path():
        return f"File: {get_signal_data_files_s2_path()}"

    def get_data_browser():
        data_browser = DataBrowser()
        if HYPYP_NIRS_DATA_PATH != "":
            data_browser.add_source(HYPYP_NIRS_DATA_PATH)
        return data_browser

    def get_preprocessor():
        value = input.recording_preprocessor()
        if value == 'upstream':
            return MnePreprocessorAsIs()
        if value == 'mne':
            return MnePreprocessorRawToHaemo()
        if value == 'cedalion':
            return CedalionPreprocessor()
        raise RuntimeError(f'Unknown preprocessor type "{value}"')
        

    @reactive.calc()
    def get_recording_s1():
        return Recording().load_file(get_signal_data_files_s1_path(), get_preprocessor())

    @reactive.calc()
    def get_recording_s2():
        return Recording().load_file(get_signal_data_files_s2_path(), get_preprocessor())

    @reactive.calc()
    def get_recording_s1_step():
        return get_recording_s1().get_preprocess_step(input.signal_data_files_analysis_property())

    @reactive.calc()
    def get_recording_s2_step():
        return get_recording_s2().get_preprocess_step(input.signal_data_files_analysis_property())

    def get_mne_raw_plot_kwargs(step, duration_range):
        try:
            range_start, range_end = duration_range
            start = range_start
            duration = (range_end - range_start)
            return dict(
                n_channels=len(step.ch_names),
                start=start,
                duration=duration,
                scalings='auto',
                show_scrollbars=False,
                block=True,
                show=False,
            )
        except:
            # TODO this is here because of CedalionStep. This code flow is ugly
            return dict()

    def mne_figure_as_image(fig):
        temp_img_path = tempfile.NamedTemporaryFile(suffix=".png").name
        fig.savefig(temp_img_path)
        return {"src": temp_img_path, "alt": "MNE Plot"}

    @render.ui
    def ui_preprocess_steps():
        # Need to wrap the plot function to have dynamic display in shiny.
        # Because of order of execution, this cannot be directly in the loop
        def bind_plot_mne_figure(step: int):
            def plot_mne_figure():
                return mne_figure_as_image(step.plot(**get_mne_raw_plot_kwargs(step, input.input_s1_mne_duration_slider())))
            # need to rename the function because every "output plot" must have a unique name
            plot_mne_figure.__name__ = f'{plot_mne_figure.__name__}_{step.key}'
            renderer = render.image(plot_mne_figure)
            # This is needed to avoid having the scrollbar on the right
            renderer._auto_output_ui_kwargs = dict(height=f'{DEFAULT_PLOT_MNE_HEIGHT}px')
            return renderer

        nav_panels = []
        recording = get_recording_s1()
        for step in recording.preprocess_steps:
            nav_panels.append(ui.nav_panel(step.desc, bind_plot_mne_figure(step)))

        return ui.navset_card_tab(*nav_panels, selected=recording.preprocess_steps[-1].desc)

    @render.plot
    def plot_s1_mne_sci():
        recording = get_recording_s1()
        fig, ax = plt.subplots()
        ax.hist(recording.quality_sci)
        ax.set(xlabel='Scalp Coupling Index', ylabel='Count', xlim=[0, 1])
        ax.title.set_text('Scalp Coupling Index')
        return fig

    @render.ui
    def ui_input_signal_choice():
        choices = []
        if input.signal_type() == 'testing':
            choices.append(ui.input_select(
                "signal_testing_choice",
                "",
                choices={
                    'chirp': 'Crossing chirps',
                    'chirp_sinusoid': 'One sinusoid, one chirp',
                    'sinusoid': 'Sinusoid different',
                    'sinusoid_almost_similar': 'Sinusoid almost similar',
                    'sinusoid_dephased': 'Sinusoid dephased',
                },
            ))
        
        elif input.signal_type() == 'data_files':
            browser = get_data_browser()
            browser.download_demo_dataset()

            choices.append(ui_option_row("Subject 1 file", ui.input_select(
                "signal_data_files_s1_path",
                "",
                choices=browser.list_all_files(),
                # for comparison with jupyter notebook
                #selected="/home/patrice/work/ppsp/HyPyP-synchro/data/NIRS/downloads/fathers/FCS28/parent/NIRS-2019-11-10_003.hdr",
            )))
            choices.append(ui_option_row("Subject 2 file", ui.input_select(
                "signal_data_files_s2_path",
                "",
                choices=[STR_SAME_AS_SUBJECT_1] + browser.list_all_files(),
            )))
            choices.append(ui_option_row("Preprocessor Class", ui.input_select(
                "recording_preprocessor",
                "",
                choices={
                    'upstream': 'MNE (no preprocessing, load as-is)',
                    'mne': 'MNE (basic fNIRS preprocessing)',
                    'cedalion': 'Cedalion (proof of concept preprocessing)',
                }
            )))
        
        return choices

    @render.ui
    def ui_input_signal_choice_property():
        options = []
        if input.signal_type() == 'data_files':
            options.append(ui_option_row(
                "Which step to use for analysis",
                ui.input_select(
                    "signal_data_files_analysis_property",
                    "",
                    choices=get_recording_s1().preprocess_step_choices,
                )
            ))
        return options

    @render.ui
    def ui_input_signal_options():
        options = []
        if input.signal_type() == 'testing':
            if input.signal_testing_choice() == 'sinusoid':
                options.append(ui_option_row("Freq 1", ui.input_numeric("signal_sinusoid_freq1", "", value=1))),
                options.append(ui_option_row("Freq 2", ui.input_numeric("signal_sinusoid_freq2", "", value=0.8))),
            elif input.signal_testing_choice() == 'sinusoid_almost_similar':
                options.append(ui_option_row("Freq", ui.input_numeric("signal_sinusoid_almost_similar_freq1", "", value=0.2))),
            elif input.signal_testing_choice() == 'sinusoid_dephased':
                options.append(ui_option_row("Freq", ui.input_numeric("signal_sinusoid_dephased_freq1", "", value=0.02))),
            elif input.signal_testing_choice() == 'chirp':
                options.append(ui_option_row("Freq Chirp from", ui.input_numeric("signal_chirp_freq1", "", value=0.2))),
                options.append(ui_option_row("Freq Chirp to", ui.input_numeric("signal_chirp_freq2", "", value=0.05))),
            elif input.signal_testing_choice() == 'chirp_sinusoid':
                options.append(ui_option_row("Freq Sinusoid", ui.input_numeric("signal_chirp_sinusoid_freq1", "", value=1))),
                options.append(ui_option_row("Freq Chirp from", ui.input_numeric("signal_chirp_sinusoid_freq2", "", value=0.2))),
                options.append(ui_option_row("Freq Chirp to", ui.input_numeric("signal_chirp_sinusoid_freq3", "", value=2))),
        
            options.append(ui_option_row("Sampling freq. (Hz)", ui.input_numeric("signal_sampling_frequency", "", value=5))),
            options.append(ui_option_row("Nb. points", ui.input_numeric("signal_n", "", value=2000))),
            options.append(ui_option_row("Noise level", ui.input_numeric("signal_noise_level", "", value=0.1))),
        
        if input.signal_type() == 'data_files':
            # TODO this try-except is here to have a PoC of Cedalion integration
            try:
                ch_names1 = get_recording_s1().preprocessed.ch_names
                ch_names2 = get_recording_s2().preprocessed.ch_names
            except:
                ch_names1 = []
                ch_names2 = []

            options.append(ui_option_row(
                "Subject 1 channel",
                ui.input_select(
                    "signal_data_files_s1_channel",
                    "",
                    choices=ch_names1
                )
                
            ))
            options.append(ui_option_row(
                "Subject 2 channel",
                ui.input_select(
                    "signal_data_files_s2_channel",
                    "",
                    choices=ch_names2
                )
            ))

        return options

    @render.ui
    def ui_input_signal_slider():
        duration = int(get_signal_duration())
        default_duration = min([DEFAULT_SIGNAL_DURATION_SELECTION, duration])
        return ui.input_slider("signal_range", "", min=0, max=duration, value=[0, default_duration], width='100%')

    @render.ui
    def ui_input_s1_mne_duration_slider():
        duration = int(get_signal_duration())
        default_duration = min([DEFAULT_SIGNAL_DURATION_SELECTION, duration])
        return ui.input_slider("input_s1_mne_duration_slider", "", min=0, max=duration, value=[0, default_duration], width='100%')

    @render.plot(height=DEFAULT_PLOT_SIGNAL_HEIGHT)
    def plot_signals():
        fig, ax = plt.subplots()
        pair = get_signals()
        offset = 0
        if input.display_signal_offset_y():
            offset = np.mean(np.abs(pair.y1)) + np.mean(np.abs(pair.y2))
        ax.plot(pair.x, pair.y1 + offset)
        ax.plot(pair.x, pair.y2 - offset)
        return fig
    

    @render.plot(height=DEFAULT_PLOT_SIGNAL_HEIGHT)
    def plot_signals_spectrum():
        fig, ax = plt.subplots()
        pair = get_signals()
        N = pair.n

        yf = fft.fft(pair.y1)
        xf = fft.fftfreq(N, pair.dt)[:N//2]
        yf2 = fft.fft(pair.y2)

        ax.plot(xf, 2.0/N * np.abs(yf[0:N//2]))
        ax.plot(xf, 2.0/N * np.abs(yf2[0:N//2]))
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.title.set_text('Spectrum')
        ax.grid()
        return fig

    @render.plot()
    def plot_mother_wavelet():
        fig, ax = plt.subplots()
        wav = get_wavelet()
        ax.plot(wav.psi_x, np.real(wav.psi))
        ax.plot(wav.psi_x, np.imag(wav.psi))
        ax.plot(wav.psi_x, np.abs(wav.psi))
        ax.title.set_text(f"mother wavelet ({wav.wavelet_name_with_args})")
        ax.legend(['real', 'imag', 'abs'])
        return fig

    @render.plot()
    def plot_wtc():
        fig, ax = plt.subplots()
        compute_coherence().plot(
            ax=ax,
            show_colorbar=False,
            show_cone_of_influence=input.display_show_coi(),
            show_nyquist=input.display_show_nyquist(),
        )
        return fig

    @render.ui
    def ui_input_wavelet_type():
        options = []
        if input.wavelet_library() == 'pywavelets':
            options.append(ui.input_select(
                "wavelet_name",
                "",
                choices={
                    'cmor': 'Complex Morlet',
                    'cgau': 'Complex Gaussian',
                },
            ))
        return options


    @render.ui
    def ui_input_wavelet_options():
        options = []
        if input.wavelet_library() == 'pywavelets':
            if input.wavelet_name() == 'cmor':
                options.append(ui_option_row("Bandwidth/Center Freq.", ui.row(
                    ui.column(6, ui.input_numeric("wavelet_bandwidth", "", value=ComplexMorletWavelet.default_bandwidth_frequency)),
                    ui.column(6, ui.input_numeric("wavelet_center_frequency", "", value=ComplexMorletWavelet.default_center_frequency)),
                )))
            if input.wavelet_name() == 'cgau':
                options.append(ui_option_row("Gaussian degree", ui.row(
                   ui.input_select(
                        "wavelet_degree",
                        "",
                        choices=list(range(1,9)),
                        selected=ComplexGaussianWavelet.default_degree,
                    ))))

        options.append(ui_option_row("Period range", ui.row(
            ui.column(6, ui.input_numeric("wavelet_period_range_low", "", value=BaseWavelet.default_period_range[0])),
            ui.column(6, ui.input_numeric("wavelet_period_range_high", "", value=BaseWavelet.default_period_range[1])),
        )))

        return options
    
    @render.ui
    def ui_input_coherence_options():
        options = []
        if input.wavelet_library() in ['pycwt']:
            pass
        return options
    
app = App(app_ui, server)


from typing import List
from pathlib import Path
import re

import numpy as np
import mne
import itertools as itertools
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist

from .preprocessor.base_preprocessor import BasePreprocessor, BaseStep
from .preprocessor.implementations.mne_preprocessor_as_is import MnePreprocessorAsIs
from .channel_roi import ChannelROI
from ..utils import (
    epochs_from_tasks,
    TASK_BEGINNING,
    TASK_END,
    TASK_NAME_WHOLE_RECORD,
    Task,
    TaskList,
    generate_random_label,
    key_dict_as_data_frame,
)
from ..wavelet.cwt import CWT
from ..wavelet.wtc import WTC

class Recording:
    """
    The Recording object encapsulates the logic around the recording for one participant.
    The preprocessing is run on the channels of the Recording

    Args:
        label (str, optional): unique label for the subject. Defaults to a random string.
        tasks (TaskList, optional): list of tasks during the recording of the participant, that will be extracted from events in the raw files to build epochs. Defaults to [].
        channel_roi (ChannelROI | None, optional): region of interest object to group channels. Defaults to None.
    """
    filepath: str | None
    subject_label: str
    channel_roi: ChannelROI | None
    mne_raw: mne.io.Raw | None
    intra_wtcs: List[WTC] | None # intra-subject wtc
    epochs_per_task: List[mne.Epochs] | None
    tasks: TaskList
    preprocess_steps: List[BaseStep] | None

    def __init__(
        self,
        subject_label:str='',
        tasks:TaskList=[],
        channel_roi:ChannelROI|None=None
    ):
        self.filepath = None
        self.subject_label = subject_label
        self.channel_roi = channel_roi
        self.mne_raw = None
        self.intra_wtcs = None
        self.epochs_per_task = None
        self.preprocess_steps = None
        self.tasks = tasks

        if len(tasks) == 0:
            # Use tasks with special values instead of time_range,
            # since we don't know yet the duration of the record
            self.tasks = [Task(TASK_NAME_WHOLE_RECORD, onset_event_id=TASK_BEGINNING, offset_event_id=TASK_END)]

    def _assert_is_preprocessed(self):
        if not self.is_preprocessed:
            raise RuntimeError('Recording is not preprocessed. Did you run preprocess() ?')

    def _assert_is_epochs_loaded(self):
        if not self.is_epochs_loaded:
            raise RuntimeError('Recording does not have epochs loaded. Did you run populate_epochs_from_tasks() ?')

    def _assert_is_wtcs_computed(self):
        if not self.intra_wtcs:
            raise RuntimeError('Recording has no intra_wtcs. Did you run compute_wtcs(with_intra=True) ?')

    @staticmethod
    def get_default_preprocessor():
        return MnePreprocessorAsIs()
    
    @property
    def task_keys(self):
        return [task.name for task in self.tasks]

    @property
    def is_preprocessed(self) -> bool:
        return self.preprocess_steps is not None
    
    @property
    def is_epochs_loaded(self) -> bool:
        return self.epochs_per_task is not None
    
    @property
    def is_wtc_computed(self):
        return self.intra_wtcs is not None

    @property
    def raw(self):
        return self.mne_raw

    @property
    def preprocessed(self) -> mne.io.Raw:
        self._assert_is_preprocessed()
        # We want the last step of all the preprocessing
        return self.preprocess_steps[-1].obj
    
    @property
    def cwts(self) -> List[CWT]:
        self._assert_is_wtcs_computed()
        seen = []
        for wtc in self.intra_wtcs:
            if wtc.cwt1 not in seen:
                seen.append(wtc.cwt1)
        return seen


    @property
    def pre(self) -> mne.io.Raw:
        return self.preprocessed
    
    @property
    def mne_preprocessed(self) -> mne.io.Raw:
        return self.preprocessed

    @property
    def mne_pre(self) -> mne.io.Raw:
        return self.pre

    @property
    def ordered_ch_names(self) -> List[str]:
        if self.channel_roi is None:
            return self.preprocessed.ch_names
        return self.channel_roi.get_ch_names_in_order(self.preprocessed.ch_names)

    @property
    def ordered_roi_names(self) -> List[str]:
        if self.channel_roi is None:
            return []
        return list(self.channel_roi.rois.keys())

    @property
    def preprocess_step_keys(self):
        self._assert_is_preprocessed()
        # get in the reverse order so that the last step is first in list
        keys = []
        for i in range(len(self.preprocess_steps)-1, -1, -1):
            keys.append(self.preprocess_steps[i].key)
        return keys

    @property
    def preprocess_step_choices(self):
        self._assert_is_preprocessed()
        steps_dict = dict()
        for i in range(len(self.preprocess_steps)-1, -1, -1):
            step = self.preprocess_steps[i]
            steps_dict[step.key] = step.desc
        return steps_dict
    
    def _fill_subject_label(self):
        if self.subject_label == '':
            self.subject_label = self.mne_raw.info['subject_info']['his_id']
        
        if self.subject_label == '':
            self.subject_label = generate_random_label(10)
    
    def load_file(
        self,
        filepath:str,
        preprocessor:BasePreprocessor|None=None,
        preprocess=True,
        verbose=False
    ):
        """
        Load a raw NIRS file as the recording of the Recording

        Args:
            filepath (str): disk path of the file to load
            preprocessor (BasePreprocess, optional): which preprocessor to use. Defaults to MnePreprocessUpstream which only load the data as-is.
            preprocess (bool, optional): if the preprocessing should be done right away, or if the file should only be loaded and the preprocessing would be run later. Defaults to True.
            verbose (bool, optional): verbosity flag. Defaults to False.

        Raises:
            RuntimeError: When file is not found

        Returns:
            self: the Recording object itself. Useful for chaining operations
        """
        if preprocessor is None:
            preprocessor = Recording.get_default_preprocessor()

        if not Path(filepath).is_file():
            raise RuntimeError(f'Cannot find file {filepath}')

        self.filepath = filepath        
        self.mne_raw = preprocessor.read_file(filepath, verbose=verbose)

        self._fill_subject_label()

        if preprocess:
            self.preprocess(preprocessor, verbose=verbose)

        return self
    
    def load_raw(
        self,
        raw:mne.io.Raw,
        preprocessor:BasePreprocessor|None=None,
        preprocess=True,
        verbose=False
    ):
        """
        Use an existing Raw object to load the data into a Recording

        Args:
            raw (mne.io.Raw): preloaded mne.io.Raw object
            preprocessor (BasePreprocess, optional): which preprocessor to use. Defaults to MnePreprocessUpstream which only load the data as-is.
            preprocess (bool, optional): if the preprocessing should be done right away, or if the file should only be loaded and the preprocessing would be run later. Defaults to True.
            verbose (bool, optional): verbosity flag. Defaults to False.

        Returns:
            self: the Recording object itself. Useful for chaining operations
        """
        if preprocessor is None:
            preprocessor = Recording.get_default_preprocessor()

        self.mne_raw = raw
        self._fill_subject_label()

        if preprocess:
            self.preprocess(preprocessor, verbose=verbose)

        return self
    
    def preprocess(self, preprocessor:BasePreprocessor, verbose:bool=False):
        """
        Run the preprocessing for the raw recording of the Recording

        Args:
            preprocessor (BasePreprocessor): which preprocessor object to use. See existing implementations or extend BasePreprocessor abstract class to develop your own
            verbose (bool, optional): verbosity flag. Defaults to False.

        Returns:
            self: the Recording object itself. Useful for chaining operations
        """
        self.preprocess_steps = preprocessor.run(self.mne_raw, verbose=verbose)
        self.populate_epochs_from_tasks(verbose=verbose)
        return self

    def get_preprocess_step(self, key:str) -> BaseStep:
        """
        Get a specific step of the preprocessing pipeline

        Args:
            key (str): name of the step

        Returns:
            BaseStep: an implementation of BaseStep class, depending on the preprocessor implementation
        """
        self._assert_is_preprocessed()
        for step in self.preprocess_steps:
            if step.key == key:
                return step

        raise RuntimeError(f'No preprocess step named "{key}"')
    
    def populate_epochs_from_tasks(self, verbose:bool=False):
        """
        Given the list of tasks (annotations and timed) of the Recording,
        find the given data in the preprocessed channels and load as epochs that will be compared between subjects
        and store them in the object

        Args:
            verbose (bool, optional): verbosity flag. Defaults to False.

        Returns:
            self: the Recording object itself. Useful for chaining operations
        """
        self.epochs_per_task = epochs_from_tasks(self.preprocessed, self.tasks, verbose=verbose)
        return self
    
    def get_epochs_for_task(self, task_name:str) -> List[mne.Epochs]:
        """
        Given a task name, return all the Epochs object

        Args:
            task_name (str): name of the task

        Returns:
            List[mne.Epochs]: list of the Epochs. Epochs must have been populated beforehand
        """
        self._assert_is_epochs_loaded()
        id = None
        try:
            id = [i for i in range(len(self.task_keys)) if self.task_keys[i] == task_name][0]
        except:
            pass

        if id is None:
            raise RuntimeError(f'Cannot find epochs for task "{task_name}"')

        return self.epochs_per_task[id]
    
    def get_roi_from_channel(self, ch_name:str) -> str:
        """
        Given a channel name, return the region of interest it belongs to, if any

        Args:
            ch_name (str): name of the channel

        Returns:
            str: name of the region of interest
        """
        if self.channel_roi is None:
            return ''
        return self.channel_roi.get_roi_from_channel(ch_name)
    
    def get_channel_to_standard_montage_map(self, as_data_frame:bool=True):
        my_montage = self.mne_raw.get_montage()
        my_positions_dict = my_montage.get_positions()['ch_pos']

        # See all montages with mne.channels.get_builtin_montages()
        # print(mne.channels.get_builtin_montages())
        standard_montage = mne.channels.make_standard_montage('standard_1020')
        standard_positions_dict = standard_montage.get_positions()['ch_pos']
        standard_keys = list(standard_positions_dict.keys())
        standard_positions = list(standard_positions_dict.values())

        my_positions_lookup = {}

        # copy source detector dict
        for k in my_positions_dict.keys():
            my_positions_lookup[k] = my_positions_dict[k]

        # add the channels which are the midpoint of source and detector
        for ch_name in self.mne_raw.ch_names:
            res = re.search(r"^(S\d+)_(D\d+).*$", ch_name)
            source = res[1]
            detector = res[2]

            ch_key = f"{source}_{detector}"
            ch_pos = (my_positions_dict[source] + my_positions_dict[detector]) / 2
            my_positions_lookup[ch_key] = ch_pos

        mapped_dict = {}

        # make sure we keep the same ordering

        # Find all distances, and nearest in standard
        for k in my_positions_lookup.keys():
            ch_pos_2d = my_positions_lookup[k].reshape(1, -1)  # reshape to 2D
            
            # Get nearest standard electrode
            dists = cdist(ch_pos_2d, standard_positions)
            
            nearest_idx = np.argmin(dists)
            mapped_dict[k] = {
                'name': standard_keys[nearest_idx],
                'dist': dists[0, nearest_idx],
            }

        if as_data_frame:
            return key_dict_as_data_frame(mapped_dict, col_names=['Source/Detector', 'Standard 1020', 'Distance'])

        return mapped_dict

    def plot_steps_for_channel(self, ch_name: str, show_cwt=True):
        # get only the "S1_D1" part of "S1_D1 760"
        ch_base_name = re.sub(' .*$', '', ch_name) 
        rows = len(self.preprocess_steps)
        if show_cwt:
            rows = rows + 1

        height = rows * 3
        fig, axes = plt.subplots(rows, 1, figsize=(12, height), sharex=True, sharey=False)
        axes = np.atleast_1d(axes)

        for j, step in enumerate(self.preprocess_steps):
            raw = step.obj
            raw_data = raw.get_data()

            for k, local_ch_name in enumerate(raw.ch_names):
                if local_ch_name == f"{ch_base_name} 760":
                    axes[j].plot(raw.times, raw_data[k,:], color='g', label=local_ch_name)
                if local_ch_name == f"{ch_base_name} 850":
                    axes[j].plot(raw.times, raw_data[k,:], color='C1', label=local_ch_name)
                if local_ch_name == f"{ch_base_name} hbo":
                    axes[j].plot(raw.times, raw_data[k,:], color='r', label=local_ch_name)
                if local_ch_name == f"{ch_base_name} hbr":
                    axes[j].plot(raw.times, raw_data[k,:], color='b', label=local_ch_name)
                axes[j].legend()
                axes[j].set_title(step.desc)

        found_cwt = None
        if show_cwt:
            for cwt in self.cwts:
                cwt_base_name = re.sub(' .*$', '', cwt.label) 
                if cwt_base_name == ch_base_name:
                    found_cwt = cwt
                    break
            if found_cwt is None:
                raise ValueError(f"Cannot find CWT for ch_name '{ch_name}'")

            cwt.plot(ax=axes[rows-1], show_colorbar=False)

        fig.suptitle(f"Steps for {ch_name}")
        fig.tight_layout()

        return fig





from itertools import compress

import mne

from ..base_step import *
from .mne_step import MneStep
from .mne_preprocessor_as_is import MnePreprocessorAsIs

class MnePreprocessorRawToHaemo(MnePreprocessorAsIs):
    """
    The MnePreprocessorRawToHaemo class uses mne fNIRS features to run some basic preprocessing steps.
    
    It does these:

    1. Store the raw data as a step
    2. Convert raw data to optical density
    3. Remove bad channels scalp_coupling_index
    4. Convert optical density to haemoglobin concentration.
    5. Filters the haemoglobin concentration based on standard high pass and low pass filters

    The steps can then be inspected on a recording, for validation.

    Use this preprocessor to explore raw data.

    NOTE: If the data loaded with MNE contains channels with `hbo` or `hbr`, it will be considered
    already preprocessed and will be returned as-is.

    """
    def __init__(self, ppf=6.0, hp_filter=0.02, lp_filter=0.7, quality_sci_threshold=0.1):
        super().__init__()
        self.ppf = ppf
        self.hp_filter = hp_filter
        self.lp_filter = lp_filter
        self.quality_sci_threshold = quality_sci_threshold
    
    def run(self, raw:mne.io.Raw, verbose:bool=False) -> list[MneStep]:
        # If have hbo or hbr, it means it is already in concentration
        if len(mne.pick_types(raw.info, fnirs=['hbo', 'hbr'])) > 0:
            raise ValueError('Loaded data seems to already be in concentrations. Use MnePreprocessorAsIs instead.')

        if verbose:
            print('Using MnePreprocessorRawToHaemo, converting fNIRS raw data to haemoglobin concentrations')

        steps = []
        steps.append(MneStep(raw, PREPROCESS_STEP_BASE_KEY, PREPROCESS_STEP_BASE_DESC))

        raw_od = mne.preprocessing.nirs.optical_density(raw)
        steps.append(MneStep(raw_od, PREPROCESS_STEP_OD_KEY, PREPROCESS_STEP_OD_DESC))

        quality_sci = mne.preprocessing.nirs.scalp_coupling_index(raw_od)
        raw_od.info['bads'] = list(compress(raw_od.ch_names, quality_sci < self.quality_sci_threshold))
        picks = mne.pick_types(raw_od.info, fnirs=True, exclude='bads')
        raw_od_clean = raw_od.copy().pick(picks)
        steps.append(MneStep(raw_od_clean, PREPROCESS_STEP_OD_CLEAN_KEY, PREPROCESS_STEP_OD_CLEAN_DESC))

        # For partial pathlength, see https://doi.org/10.1117/1.JBO.18.10.105004, Table 1
        raw_haemo: mne.io.Raw = mne.preprocessing.nirs.beer_lambert_law(raw_od_clean.copy(), ppf=self.ppf)
        steps.append(MneStep(raw_haemo, PREPROCESS_STEP_HAEMO_KEY, PREPROCESS_STEP_HAEMO_DESC))

        raw_haemo_filtered = raw_haemo.copy().filter(self.hp_filter, self.lp_filter, verbose=verbose, method='iir')
        steps.append(MneStep(raw_haemo_filtered, PREPROCESS_STEP_HAEMO_FILTERED_KEY, PREPROCESS_STEP_HAEMO_FILTERED_DESC))

        return steps
        